#include "SDLFont.h"
#include <assert.h>


#define CHARACTERS 256

FT_Library library = NULL;

SDLFace* SDLFace_load(char* path, Uint32 size)
{
	FT_Face font		= NULL;

	if (library == NULL)
	{
		int result = FT_Init_FreeType( &library );
		if ( result )
		{
			fprintf(stderr, "SDLFont.c: Unable to initialize freetype library. Error: %i\n", result);
			return NULL;
		}
	}

	int result = FT_New_Face(library, path, 0, &font);
	if ( result )
	{
		fprintf(stderr, "SDLFont.c: Unable to load font '%s', check your font paths. Error code: %i\n", path, result);
		return NULL;
	}

	if (FT_Set_Char_Size(font, 0, size*64, 58, 58))
	{
		fprintf(stderr, "SDLFont.c: Unable to set font '%s' to size: %i\n", path, size);
		return NULL;
	}

	SDLFace* returnValue = SDLFace_create(font);

	return returnValue;

}

SDLFace* SDLFace_create(FT_Face face)
{
	SDLFace* sdlFace = (SDLFace*) malloc( sizeof(SDLFace) );

	/* Create the initial white version */
	SDLFont* white = SDLFont_create(face, 255, 255, 255);

	sdlFace->fonts = Vector_construct();
	Vector_add(sdlFace->fonts, white);

	return sdlFace;
}

SDLFont* SDLFace_getColour( SDLFace* face, Uint32 colour )
{
	int i 			= 0;
	int size 		= 0;
	SDLFont* current	= NULL;


	size = Vector_getSize(face->fonts);

	for (i = 0; i < size; i++)
	{
		current = (SDLFont*)Vector_get(face->fonts, i);

		if (current->colour == colour)
		{
			//If this font is not first, swap so that it is - as the same colour is often requested several times
			if (i != 0)
			{
				SDLFont* temp = (SDLFont*) Vector_get(face->fonts, 0);
				Vector_set(face->fonts, 0, current);
				Vector_set(face->fonts, i, temp);
			}
			return current;
		}
	}

	//The colour we are looking for does not exist - so create it.
	current = SDLFont_createFrom( (SDLFont*)Vector_get(face->fonts, 0), colour );
	Vector_add(face->fonts, current);

	return current;
}

SDLFont* SDLFont_createFrom( SDLFont* font, Uint32 colour)
{
	int i = 0;
	SDLFont* newFont = (SDLFont*) malloc( sizeof(SDLFont) );

	newFont->colour 	= colour;
	newFont->ascender 	= font->ascender;
	newFont->descender 	= font->descender;

	for (i = 0; i < CHARACTERS; i++)
	{
		newFont->characters[i] = SDLCharacter_copy(font->characters[i], colour);
	}
	return newFont;
}

SDLFont* SDLFont_create(FT_Face face, Uint8 red, Uint8 green, Uint8 blue)
{
	int i = 0;
	SDLFont* newFont = (SDLFont*) malloc( sizeof(SDLFont) );

	for (i = 0; i < CHARACTERS; i++)
	{
		newFont->characters[i] = SDLCharacter_create(face, i, red, green, blue);
	}
	newFont->colour 	= SDL_MapRGB(newFont->characters[0]->bitmap->format, red, green, blue);
	newFont->ascender 	= FT_MulFix(face->ascender, face->size->metrics.y_scale)>>6;
	newFont->descender 	= - ((FT_MulFix(face->descender, face->size->metrics.y_scale))>>6);

	return newFont;
}

SDLCharacter* SDLCharacter_create(FT_Face face, char character, Uint8 red, Uint8 green, Uint8 blue)
{
	SDLCharacter* newChar 	= NULL;
	int i 			= 0;
	int j 			= 0;
	FT_Bitmap* bitmap 	= NULL;
	Uint8* pixels		= NULL;
	SDL_Surface* surface	= NULL;
	Uint32* pixel		= NULL;
	SDL_PixelFormat* format = NULL;
	Uint8 alpha 		= 0;


	//Make the freetype system render the font
	int glyph_index = FT_Get_Char_Index( face, (unsigned char)character );

	FT_Load_Glyph( face, glyph_index, FT_LOAD_DEFAULT );
	FT_Render_Glyph( face->glyph, FT_RENDER_MODE_NORMAL );
	bitmap = &(face->glyph->bitmap);

	//Convert it to a SDL_Surface
	newChar = (SDLCharacter*) malloc( sizeof(SDLCharacter) );

	surface = SDL_CreateRGBSurface(SDL_SWSURFACE, bitmap->width, bitmap->rows, 32, 0xff000000, 0x00ff0000, 0x0000ff00, 0x000000ff);

	newChar->bitmap = SDL_DisplayFormatAlpha(surface);
	SDL_FreeSurface(surface);
	surface = newChar->bitmap;

	pixels = (Uint8*)surface->pixels;
	format = surface->format;

	for (i = 0 ; i < bitmap->rows; i++)
	{
		pixel = (Uint32*)pixels;

		for (j = 0; j < bitmap->width; j++)
		{
			alpha = bitmap->buffer[i*bitmap->pitch+j];

			*pixel = SDL_MapRGBA(format, red, green, blue, alpha);
			pixel++;
		}
		pixels += surface->pitch;
	}

	pixels = (Uint8*)surface->pixels;
	pixel = (Uint32*)pixels;

	//Calculate various metrics
	newChar->xDisplace 	= face->glyph->bitmap_left;
	newChar->yDisplace 	= face->glyph->bitmap_top;
	newChar->advance 	= face->glyph->advance.x >> 6;

	return newChar;
}

SDLCharacter* SDLCharacter_copy( SDLCharacter* character, Uint32 colour)
{
	int i 				= 0;
	int j 				= 0;
	Uint8 red 			= 0;
	Uint8 green 			= 0;
	Uint8 blue 			= 0;
	Uint8 newRed 			= colour>>16;
	Uint8 newGreen 			= colour>>8;
	Uint8 newBlue 			= colour;
	Uint8 alpha			= 0;
	Uint32* src 			= NULL;
	Uint32* dst 			= NULL;
	Uint8* srcLine 			= NULL;
	Uint8* dstLine 			= NULL;
	SDL_Surface* dstSurface 	= NULL;
	SDL_Surface* srcSurface		= character->bitmap;

	SDLCharacter* copy 	= (SDLCharacter*) malloc( sizeof(SDLCharacter) );

	copy->xDisplace 	= character->xDisplace;
	copy->yDisplace 	= character->yDisplace;
	copy->advance		= character->advance;
	dstSurface		= SDL_CreateRGBSurface(SDL_SWSURFACE, srcSurface->w, srcSurface->h, 32, 0xff000000, 0x00ff0000, 0x0000ff00, 0x000000ff);
	copy->bitmap 		= SDL_DisplayFormatAlpha(dstSurface);
	SDL_FreeSurface(dstSurface);
	dstSurface 		= copy->bitmap;

	srcLine = (Uint8*)srcSurface->pixels;
	dstLine = (Uint8*)dstSurface->pixels;

	for (i = 0; i < srcSurface->h; i++)
	{
		src = (Uint32*)srcLine;
		dst = (Uint32*)dstLine;

		for (j = 0; j < srcSurface->w; j++)
		{
			SDL_GetRGBA(*src, srcSurface->format, &red, &green, &blue, &alpha);
			*dst = SDL_MapRGBA(dstSurface->format, newRed, newGreen, newBlue, alpha);

			src++;
			dst++;
		}
		srcLine += srcSurface->pitch;
		dstLine += dstSurface->pitch;
	}

	return copy;
}

SDLCharacter* SDLFont_getCharacter( SDLFont* font, char character )
{
	return font->characters[(Uint8)character];
}

SDLFont* SDLFace_getAny( SDLFace* face)
{
	return (SDLFont*) Vector_get(face->fonts, 0);
}

Uint32 SDLFont_getWidth(SDLFont* font, char* string)
{
	int i 			= 0;
	Uint32 width		= 0;
	SDLCharacter* chr	= NULL;

	for (i = 0; string[i] != 0; i++)
	{
		chr = SDLFont_getCharacter( font, string[i] );
		width += chr->advance;
	}
	return width;
}

Uint32 SDLFont_getWidthSubstring(SDLFont* font, char* string, Uint32 from, Uint32 to)
{
	Uint32 i;
	Uint32 pixelWidth = 0;
	SDLCharacter* chr	= NULL;

	for (i = from; i < to; i++)
	{
		chr = SDLFont_getCharacter( font, string[i] );
		pixelWidth += chr->advance;
	}
	return pixelWidth;
}



Uint32 SDLFont_draw(SDLFont* font, char* string, Uint32 textLength, SDL_Surface* surface, Uint32 x, Uint32 y)
{
	Uint32 width		= 0;
	SDLCharacter* chr	= NULL;
	SDL_Rect dstrect;
	SDL_Rect srcrect;


	srcrect.x = 0;
	srcrect.y = 0;

for (Uint32 i = 0; i < textLength; i++)
	{
		chr = SDLFont_getCharacter( font, string[i] );

		srcrect.w = chr->bitmap->w;
		srcrect.h = chr->bitmap->h;

		dstrect.x = x + chr->xDisplace;
		dstrect.y = y - chr->yDisplace + font->ascender;

		SDL_BlitSurface(chr->bitmap, &srcrect, surface, &dstrect);

		x 	+= chr->advance;
		width 	+= chr->advance;
	}

	return width;
}

