#include "SDL_System.h"

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "SDL/SDL.h"
#include "SDL/SDL_thread.h"
#include "SDLTimer.h"
#include "Message.h"
#include "Vector.h"
#include "SDL_Draw.h"
#include "SDLFont.h"
#include "ConfigPanel.h"
#include "Sample.h"
#include "GUI.h"
#include "FontMapping.h"
#include "../../core/magazine.h" //To get the version string
#include "../../core/swap.h"

using namespace SDLGUI;


char* ID_STRING = "\nPanorama - e-Magazine Browser\nVersion %s\nCopyright (c) 1999-2005 Chris Dragan. All rights reserved.\nhttp://chris.dragan.name/\n\nSDL version by Michael Barkholt (mibar02@student.sdu.dk)\n\n";

char* HELP = "\t-h, -help\t\tthis text\n\t-c, -config\t\tforce showing of configuration window\n\t-s X, -song X\t\tplay song X instead of default music\n\t-a X, -article X\tshow article X instead of default\n";

/*
   Nothing to talk about. Simply process user configuration here, and,
   of course, init your variables and stuff. This is the only non-virtual
   function - the rest you have to derive (that's obvious...). Naturally,
   you can create your own methods in your derived system class.
*/
SDLSystem::SDLSystem(Platform* _platform, int argc, char** argv):
openFile(NULL), platform(_platform), config(NULL), current(NULL), videoSurface(NULL), ready(false), messageBuffer(NULL),
postSync(NULL), cursorImage(NULL), cursorShow(false), 
fontBuffer(NULL), fontMap(NULL), fontMapping(NULL), errorMessage(NULL), searchPanel(NULL), 
searchWords(NULL), searchArticles(NULL), soundSystem(NULL), filesInDir(NULL), foundFiles(NULL), 
fileIndex(0)
{
	/* Read configuration */
	config = new Config(platform->getConfigPath(), platform);
	
	/* Parse the argument string */
	for (int i = 1; i < argc; i++)
	{
		if ( strcasecmp(argv[i], "-h") == 0 || strcasecmp(argv[i], "-help") == 0 )
		{
			printf(ID_STRING, __VERSION_STRING);
			printf(HELP);
			return;
		}
		else if ( strcasecmp(argv[i], "-c") == 0 || strcasecmp(argv[i], "-config") == 0 )
		{
			config->setShowConfig(true);
		}
		else if ( strcasecmp(argv[i], "-s") == 0 || strcasecmp(argv[i], "-song") == 0 )
		{
			if (i+1 < argc)
				config->setExternalTune(argv[i+1]);
			i++;
		}
		else if ( strcasecmp(argv[i], "-a") == 0 || strcasecmp(argv[i], "-article") == 0 )
		{
			if (i+1 < argc)
				config->setArticle(argv[i+1]);
			i++;
		}
	}
	int initFlags = SDL_INIT_VIDEO | SDL_INIT_TIMER;

	/* Initialize SDL */	
	if ( SDL_Init(initFlags) < 0 ) 
	{
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		ready = false;
	}
	else
		ready = true;
	
	messageBuffer 	= Vector_construct();
	postSync	= SDL_CreateMutex();
	messageQueue	= Queue_construct();
	fontBuffer	= Map_createMap(10);
	searchWords	= Vector_construct();
	searchArticles	= Vector_construct();
	foundFiles 	= Vector_construct();

	/* Ensure that the TEMP and SAVE directories exist */
	char* dirPath = new char[1024];
	snprintf(dirPath, 1024, "%ssave", platform->getCurrentDir());
	platform->makedir(dirPath);
	snprintf(dirPath, 1024, "%stemp", platform->getCurrentDir());
	platform->makedir(dirPath);
	delete[] dirPath;

	
	/* Find all the files in the current directory */
	filesInDir	= Vector_construct();
	platform->createDirListing( platform->getCurrentDir(), filesInDir);

	Vector* fontPaths = config->getFontPaths();
	fontMapping = new FontMapping(platform);
	
	for(int i = 0; i < Vector_getSize(fontPaths); i++)
	{
		char* nextPath = (char*) Vector_get(fontPaths, i);
		fontMapping->parse(nextPath);
	}
	fontMapping->writeToCache();
	fontMap = fontMapping->getMapping();
	
	GUI::setFontMap(fontMap);
	
	/* Initialize cursorArea memory */
	cursorArea.x = 0;
	cursorArea.y = 0;
	cursorArea.w = 0;
	cursorArea.h = 0;
	
	if (config->getVerbose())
		printf("SDLSystem constructor\n");

	
	/* Show the configuration GUI */
	if (config->getShowConfig() && ready)
	{
		Uint32 width 	= 250;
		Uint32 height 	= 207;
		bool cont 	= true;
		bool start 	= true;
		ConfigPanel* cfgPanel = new ConfigPanel(&cont, &start, config, width, height);

		videoSurface = SDL_SetVideoMode(width, height, 32, SDL_SWSURFACE);
		if ( videoSurface == NULL ) 
		{
			fprintf(stderr, "Unable to set %ix%i video: %s\n", width, height, SDL_GetError());
			ready = false;
			return;
		}
		set_window_title("Panorama 2");
	
		/* Display configuration GUI */
		SDL_Event event;
		while (cont)
		{
			cfgPanel->paint(videoSurface, 0, 0);
			SDL_Flip(videoSurface);
	
			if( SDL_PollEvent(&event))
				cfgPanel->handleEvent(&event, 0, 0);
		}	
		if (cfgPanel->getModule() != NULL)
			config->setExternalTune( cfgPanel->getModule() );
		
		config->save();
		delete cfgPanel;
		
		if (start == false)
			ready = false;
	}
	
	if (config->getMusic() || config->getSoundEffects())
		soundSystem = new DumbAndOgg(config->getMixingFrequency());

}

/* Returns true if the object is available (esp. graphics)
   This method is only for you. You can use it to test if your system
   object have been initialized. Note that if you don't need it, you
   still have to override it with an empty method, because all virtual
   methods in the System class are abstract!
*/
bool SDLSystem::is_ready()
{
	if (config->getVerbose())
		printf("is_ready()\n");
	
	return ready;	
}

/* Shuts down the system (graphics, sound, optional message) 
   Do all cleanup here. Shutdown display, sound systems, etc, etc.
*/
SDLSystem::~SDLSystem()
{
	if (config->getVerbose())
		printf("~SDLSystem\n");
}

/* On first call, the filename points to a mask (e.g. *.*). The
function is called until it returns false, what indicates that it
cannot find any matching files.
   Let's look at an example:
      char* file = "*.*";
      while ( system->find_files( file ) ) {
	 some_object->process( file );
      }
   On the first call a mask is given. The method is called until nothing
   is found (false is returned in this case). On subsequent calls the
   argument's value is indeterminate. Don't delete[] the argument before
   setting it to a found file name - this is handled externally !
   The passed argument is either a "*.*" mask or "*.ext" mask, where "ext"
   is some extension. You should assume that the current directory is
   the directory where the .dat file is located.
*/

bool mask = false;

bool SDLSystem::find_files (char* &filename)
{
	if (config->getVerbose() && mask == false)
		printf("Attempting to find files: %s\n", filename);

	if ( mask == false && strcmp("*.*", filename) == 0 )
	{
		mask = true;
		Vector_empty(foundFiles);
		for (int i = 0; i < Vector_getSize(filesInDir); i++)
			Vector_add(foundFiles, Vector_get(filesInDir, i) );
		fileIndex = 0;
	}
	
	if (fileIndex >= 0 && fileIndex < Vector_getSize(foundFiles) )
	{
		filename = (char*) Vector_get(foundFiles, fileIndex);
		fileIndex++;
		return true;
	}
	else
	{
		fileIndex = -1;
		mask = false;
		return false;
	}
}

/* Checks if a file exists 
   Similar to find_files(), but is called only once to check if a file
   exists. The argument is either a "*.ext" mask or an entire file
   name (with path).
*/
bool SDLSystem::find_one_file (char* &filename)
{
	if (config->getVerbose())
		printf("find_one_file : %s\n", filename);
	
	if (filename[0] == '*')
	{
		char* ext = filename+1;
		Uint32 extLen = strlen(ext);
		for (int i = 0; i < Vector_getSize(filesInDir); i++)
		{
			char* next = (char*) Vector_get(filesInDir, i);
			Uint32 nextLen = strlen(next);
			
			if (nextLen >= extLen)
			{
				if ( strcasecmp(next+nextLen-extLen, ext) == 0)
				{
					char* copy = new char[ strlen(next)+1 ];
					sprintf(copy, "%s", next);
					filename = copy;
					return true;
				}
			}
		}
	}
	else
	{
		FILE* file = fopen(filename, "rb");
		if (file != NULL)
		{
			fclose(file);
			return true;
		}
	}
	return false;
}

/* Creates a file. Returns false if cannot create. The handle to
the file is kept internally in the class, as only one file
is opened at a time. The "where" parameter specifies where
the file is to be created. It also has a flag telling whether
to overwrite an existing file.
   Creates a file. Do not be confused about the handle - you keep and
   manage it. The program assumes to have opened only one file at a
   time, but you can override this behavior putting files in some
   structure (just like you do with fonts - see below). Though I
   recommend keeping one file open at a time. The "where" parameter
   specifies where the file is to be created: in current directory
   (with the .dat file or with the executable - depends on you)
   in some temporary directory (e.g. specified by TEMP environment
   variable) or in SAVE directory (create a subdirectory in the directory
   with the .dat file or with the executable).
   Returns true if succeeds or false if fails (that's obvious...).
   CREATE_IN_PROGRAM_DIR        = 0,
   CREATE_IN_SAVE_DIR           = 1,
   CREATE_IN_TEMP_DIR           = 2,
   CREATE_IN_CURRENT_DIR        = 3,
   CREATE_OVERWRITE             = 4

*/
bool SDLSystem::create_file (const char* filename, UINT where)
{
	if (config->getVerbose())
		printf("create_file: %s, where: %i\n", filename, where);


	char* realPath	= (char*) malloc( strlen(filename) + 1 + 5 );
	char* flags	= NULL;
	
	if (where & CREATE_IN_PROGRAM_DIR || where & CREATE_IN_CURRENT_DIR)
		sprintf(realPath, "%s", filename);

	if (where & CREATE_IN_SAVE_DIR)
		sprintf(realPath, "save/%s", filename);
	
	if (where & CREATE_IN_TEMP_DIR)
		sprintf(realPath, "temp/%s", filename);
	
	if (where & CREATE_OVERWRITE)
		flags = "w+b";
	else
		flags = "a+b";
	
	openFile = fopen(realPath, flags);
	
	free(realPath);
	
	if (openFile)
		return true;

	return false;	
}

/* Opens a file. Returns false if cannot open. Handle as above.
   Opens a file. Both creating and opening operations assume sequential
   file access with rare pointer movements.
      Opens a file. Both creating and opening operations assume sequential
   file access with rare pointer movements. I do not use standard C
   functions, since some systems (like Win32 and BeOS) supply better ones.
*/
bool SDLSystem::open_file (const char* filename)
{
	if (config->getVerbose())
		printf("open_file: %s\n", filename);
	
	openFile = fopen(filename, "rb");
	
	if (openFile == NULL)
	{
		char* tempPath = new char[strlen(filename) + strlen("temp/") + 1];
		sprintf(tempPath, "temp/%s", filename);
		openFile = fopen(tempPath, "rb");
		delete[] tempPath;
	}
	if (openFile == NULL)
	{
		fprintf(stderr, "Could not open file '%s' for reading.\n", filename);
		return false;
	}
	return true;
}

/* Returns size of an opened file */
DWORD SDLSystem::get_file_size()
{
	if (config->getVerbose())
		printf("get_file_size: ");

	assert(openFile != NULL);
	
	DWORD size = 0;	
	long currentPos = ftell(openFile);
	
	fseek(openFile, 0, SEEK_END);
	size = ftell(openFile);
	
	fseek(openFile, currentPos, SEEK_SET);
	
	if (config->getVerbose())
		printf("%i\n", size);
	
	return size;
}

/* Sets file position. Returns false if the pointer exceedes file size
or if cannot set the position. */
bool SDLSystem::seek_file (DWORD position)
{
	if (config->getVerbose())
		printf("seek_file pos: %i\n", position);
	

	if ( fseek(openFile, position, SEEK_SET) != 0)
        {
                printf("Seek error!\n");
		return false;
        }

	return true;
}

/* Reads opened file.*/
bool SDLSystem::read_file (BUFFERPTR buf, DWORD size)
{
	if (config->getVerbose())
		printf("read_file, size: %i - ", size);
	
	unsigned int read = 0;
	
	read = fread(buf, 1, size, openFile);	
	
	if (config->getVerbose())
		printf("\n", read);

	if (read != size)
		return false;

	return true;
}

/* Writes to opened file. */
bool SDLSystem::write_file (BUFFERPTR buf, DWORD size)
{
	assert (openFile != NULL);

	if (config->getVerbose())
		printf("write_file, %i bytes\n", size);
	
	return fwrite(buf, 1, size, openFile) == size;
}

/* Closes opened file.*/
void SDLSystem::close_file()
{
	assert(openFile != NULL);
	if (config->getVerbose())
		printf("close_file\n");
	fclose(openFile);
	openFile = NULL;
}

/* Initializes window and video mode. */
bool SDLSystem::init_display (UINT width, UINT height)
{
	Uint32 flags 	= 0;
	
	if (config->getVerbose())
		printf("init_display, width: %i height: %i\n", width, height);
	
	flags = SDL_SWSURFACE;

	if (config->getFullscreen())
		flags |= SDL_FULLSCREEN;

	if (videoSurface != NULL)
		SDL_FreeSurface(videoSurface);
	
	videoSurface = SDL_SetVideoMode(width, height, 32, flags);
	if ( videoSurface == NULL ) 
	{
		fprintf(stderr, "Unable to set %ix%i video: %s\n", width, height, SDL_GetError());
		return false;
	}
	set_window_title("Panorama 2");

	/* Two software buffers for rendering */
	current = SDL_DisplayFormat(videoSurface);
	
	/* Construct the search window GUI */
	Uint32 searchWidth 	= 200;
	Uint32 searchHeight 	= 300;
	searchPanel	= new SearchPanel(searchWords, searchArticles, (width-searchWidth)/2, (height-searchHeight)/2, searchWidth, searchHeight);

	/* We will be using a software cursor, since the hardware version do not support 24 bit colours */
	SDL_ShowCursor(SDL_DISABLE);



	
	return true;
}

/* Sets title of the program window */
void SDLSystem::set_window_title (const char* title)
{
	if (config->getVerbose())
		printf("set_window_title, %s\n", title);

	SDL_WM_SetCaption(title, title);
}

/* Installs truetype font ("where" is the same as for create_file())
   This gives you a font (TTF) file. These files are saved in your
   temp directory (see create_file()) prior to calling this method,
   so "from_where" should specify temp. You don't have to care much
   about this method - just warn the users in the doc that they
   should install appropriate fonts before running the program.
   These fonts should also be supplied by container/issue authors.
*/
void SDLSystem::install_font (const char* filename, UINT from_where)
{
	if (config->getVerbose())
		printf("install_font, '%s', from: %i\n", filename, from_where);
	
	Vector* dirList = Vector_construct();
	
	char* tempDir = new char[2048];
	snprintf(tempDir, 2048, "%s%s%s", platform->getCurrentDir(), "temp", platform->getDirSeparator());
	
	platform->createDirListing( tempDir, dirList);
	delete[] tempDir;
	
	for (int i = 0; i < Vector_getSize(dirList); i++)
	{
		char* entry = (char*) Vector_get(dirList, i);
		
		if (strstr(entry, filename) == NULL)
		{
			Vector_remove(dirList, i);
			i--;
			delete[] entry;
		}
	}
	
	fontMapping->createFontMapping(dirList, fontMap);
}

/* Creates a font object. See the corresponding stuff in system.cpp
   This will be difficult for you if your system does not support
   TrueType fonts. You have to create a font according to the use specs.
   Flags are defined in system.h (bold, italic, underline). Charsets too.
   You return height=ascent and add_height=descent. Do not add the additional
   spacing. FONT is defined as void*, so you have to supply your own
   structure and cast it to (FONT). Beware of duplicating the fonts,
   the create_font() requests will be multiple, and often the parameters
   will be the same. The face will always be lowercase. If the requested
   face doesn't exist, you should use some default font (perhaps system
   default). Like in all the other cases, the Win32 and BeOS SI sources
   supplied by me should show you some tips.

   CHARSET_DEFAULT              = 0
   CHARSET_OEM                  = 1
   CHARSET_EASTEUROPE           = 2
   CHARSET_RUSSIAN              = 3
   CHARSET_TURKISH              = 4
   CHARSET_GREEK                = 5

   FF_BOLD                      = 1
   FF_ITALIC                    = 2
   FF_UNDERLINE                 = 4

*/
FONT SDLSystem::create_font (const char* faceName, UINT required_height, UINT flags, UINT charset, UINT& height, UINT& add_height)
{
	char* fontName 		= NULL;
	char* path		= NULL;
	char* face		= NULL;
	SDLFace* sdlFace 	= NULL;
	
	if (config->getVerbose())
		printf("create_font, '%s', height: %i, flags: %x, charset: %i\n", faceName, required_height, flags, charset);

	face = (char*) malloc(strlen(faceName) + 3);
	sprintf(face, "%s:%i", faceName, flags);
	
	path = (char*) Map_get(fontMap, face);
	
	if (path == NULL)
	{
                if (config->getVerbose())
		      fprintf(stderr, "Loading of unmapped font, '%s' requested. Falling back to default.\n", face);

		face = "default:0";
		path = (char*) Map_get(fontMap, face);
	}

	/* Copy the font name, since the core system reuses the memory! */
	fontName = (char*) malloc( strlen(face) + 100 ); //Oh my god, magic constant on a buffer - typical buffer overflow invitation

	sprintf(fontName, "%s:%i", face, required_height);
	
	
	/* Check if this particulare font has allready been loaded */
	sdlFace = (SDLFace*) Map_get(fontBuffer, fontName);
	
	if (sdlFace != NULL)
	{
		height		= SDLFace_getAny(sdlFace)->ascender;
		add_height	= SDLFace_getAny(sdlFace)->descender;
		return sdlFace;
	}
	sdlFace = SDLFace_load(path, required_height);
	
	height		= SDLFace_getAny(sdlFace)->ascender;
	add_height	= SDLFace_getAny(sdlFace)->descender;

	Map_put(fontBuffer, fontName, sdlFace);
	
	return sdlFace;
}


/* Returns width of a text taking into account the font.
   "font" is a font supplied by you with create_font(). Returns actual
   text width in pixels.
*/
UINT SDLSystem::get_text_width (const char* text, FONT _font)
{
	SDLFace* face		= (SDLFace*)_font;
	SDLFont* font 		= SDLFace_getAny(face);
	
	if (config->getVerbose())
		printf("get_text_width: %s\n", text);

	return SDLFont_getWidth(font, (char*)text);
}

/* Renders text. The return value is important when printing -
it's the width of the printed text, given in Panorama coordinates.
   "text_len" is provided so you do not have to call strlen(text).
   (x,y) specifies upper left corner of the rectangle within which
   the text will be drawn. Take care of the region ! Color format
   specified by you in get_user_flag().
   See print() for meaning of return value.
*/
UINT SDLSystem::out_text (const char* text, UINT text_len, UINT x, UINT y, DWORD color, FONT _font)
{
	color = SWAP24(color);

	SDLFace* face		= (SDLFace*)_font;
	SDLFont* font 		= SDLFace_getColour(face, color);

	if (config->getVerbose())
		printf("out_text, '%s', length: %i, x: %i, y: %i, color:%i\n", text, text_len, x, y, color);

	return SDLFont_draw(font, (char*) text, text_len, current, x, y);
}

/* Creates an image. pBuf points to an image in 24-bit format. The system
will probably want to convert it to a more suitable format matching
the screen mode. pMask is either NULL or specifies a 1-bit image mask
 (bit=1 - transparent, bit=0 - opaque).
   A request to create an image. The image duplication is handled within
   the engine, so you do not have to bother about it like in case of fonts.
   "buf" contains the actual image. You can specify the order of R, G and
   B components, the byteness (3 or 4 bytes per pixel) and the width
   alignment (1, 2, 4 or 8 pixels). Your requests are read through
   get_user_flag() (see below). The most trouble will supply the "mask"
   probably. It shows you which places in the image are transparent.
   The problem is, that this mask has 1-bit depth. The width of the
   mask is aligned on 32 pixels (on 4 bytes) not on 8 pixels. This
   limitation comes from the BMP format. The order of bits is: bit 7
   comes first (is leftmost) and bit 0 comes last (is rightmost).
   If a bit is set (1) the corresponding pixel in the image is transparent.
   Bit=0 indicates opaque pixel. This is perfect for using AND-OR
   technique. In this case you can assume that all transparent places in
   the image are actually black. However if you use colorkeying technique,
   you should clear those places somehow (JPEG decompression may supply
   not entirely black areas). Note that not all images are semi-transparent,
   so if you get mask==NULL, the image (buf) is entirely visible.
   To clarify: the "width" parameter contains actual, visible image width.
   The masks width is 32-pixel aligned (e.g. ((width-1)&(-32))+32) and
   the image width is aligned according to your settings (see
   get_user_flag()). Do not forget about this.
*/
IMAGE SDLSystem::create_image (UINT width, UINT height, BUFFERPTR buf, BUFFERPTR mask)
{
	Uint32* copy 		= NULL;
	Uint8* bitmask 		= (Uint8*) mask;
	Uint32 bytesPerMaskLine = 0;

	if (config->getVerbose())
		printf("create_image, width: %i, height: %i\n", width, height);

	/* Create copy since the engine later frees the buffer we have been given */	
	copy = (Uint32*) malloc ( sizeof(Uint32)* width* height);
	memcpy(copy, buf, width*height*sizeof(Uint32));
	SDL_Surface* image = NULL;

        image = SDL_CreateRGBSurfaceFrom(copy, width, height, 32, width*4, 0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000);
	/* Determine the number of bytes in each line of the mask - its a bitmask thats 32 bit aligned. */	
	bytesPerMaskLine = image->w/8;
	if (image->w%8 != 0)
		bytesPerMaskLine++;
	
	if (bytesPerMaskLine % 4 != 0)
		bytesPerMaskLine = bytesPerMaskLine + (4 - bytesPerMaskLine % 4);
	
	int i = 0;
	int j = 0;
	
	for (i = 0; i < image->w; i++)
	{
		for (j = 0; j < image->h; j++)
		{
			Uint32* pixel = ((Uint32*)(((Uint8*)image->pixels)+j*image->pitch+i*image->format->BytesPerPixel));
						
			
			if (((*pixel) & 0x00ffffff) == 0 && mask != NULL) 	//The colour black is supposed to mean transparent
				*pixel &= 0x00ffffff;
			else
				*pixel |= 0xff000000;	//Make completly opaque
				
						
			if (bitmask != NULL)
			{
				
				Uint8 maskbyte = bitmask[bytesPerMaskLine*j+i/8];
				maskbyte &= ( 128 >> (i % 8));
				
				if (maskbyte == 0)
					*pixel |= 0xff000000;
				else
					*pixel = 0;
				
			}
		}
	}
	return image;
}

/* Destroys created image.*/
void SDLSystem::unload_image (IMAGE image)
{
	if (config->getVerbose())
		printf("unload_image\n");
	
	SDL_FreeSurface((SDL_Surface*)image);
}


/* Indicates start of drawing. This method will probably lock the
drawing surface (maybe backbuffer). All the following drawing
routines (except Flip()) should be called between BeginDisplay()
and EndDisplay() - a check is appropriate. Note that also
blt( MOVIE ) routine can be called only between Begin/EndDisplay.*/
bool SDLSystem::begin_display()
{
	if (config->getVerbose())
		printf("begin_display - must lock: %i\n", SDL_MUSTLOCK(current) );
	set_region(0, 0, 0, 0);
	
	if (SDL_MUSTLOCK(current))
		SDL_LockSurface(current);

	return true;
}

/* Unlocks the drawing surface. Called after the drawing is finished.*/
void SDLSystem::end_display()
{
	if (config->getVerbose())
		printf("end_display\n");

	if (SDL_MUSTLOCK(current))
		SDL_UnlockSurface(current);
	
}

/* Shows the result of drawing on the screen. effect=0 - normal flip,
effect=1 - cross-fade (duration in millisec.). clearFlags specify
whether the front or the backbuffer is to be cleared before flipping
or fading (fade-in or fade-out).
   If effect==EFF_NONE (0) you simply do a flip or whatever. But if
   it is EFF_XFADE, you are supposed to do a cross-fading. The duration
   of the effect is in "duration" parameter. You can take advantage
   of the assembly source supplied by me, if you are on a x86 machine.
   Clear flags may request clearing the primary (visible) surface before
   fading, or clearing the drawing surface (the result of the flip will
   be black in this case).
*/
void SDLSystem::flip (UINT effect, UINT clear_flags, UINT duration)
{
	SDL_Rect screenrect;
	screenrect.x = 0;
	screenrect.y = 0;
	screenrect.w = current->w;
	screenrect.h = current->h;
	
	if (config->getVerbose())
		printf("flip, effect: %i, clear_flags: %i, duration: %i\n", effect, clear_flags, duration);

	if ( (clear_flags & CLEAR_FRONT) != 0)
		SDL_FillRect(videoSurface, NULL, 0);
	if ( (clear_flags & CLEAR_BACK) != 0)
		SDL_FillRect(current, NULL, 0);
	
	/* We have been drawing on 'current', and 'videoSurface' is the surface that represents
	the actual physical screen */

	switch (effect)
	{
		case EFF_XFADE:
			SDL_Crossfade(current, videoSurface, duration);
			if (cursorShow)
				move_cursor(cursorArea.x+cursorX, cursorArea.y+cursorY);
			break;
		case EFF_NONE:
			SDL_BlitSurface(current, &screenrect, videoSurface, &screenrect);

			if (cursorShow)
				move_cursor(cursorArea.x+cursorX, cursorArea.y+cursorY);
			else
				SDL_Flip(videoSurface);
			break;		
	}	

}

/* Erases everything from the visible surface
   What's visible becomes black.
*/
void SDLSystem::clear_display()
{
	if (config->getVerbose())
		printf("clear_display\n");
	
	SDL_FillRect(videoSurface, NULL, 0);
	SDL_Flip(videoSurface);
}

/* Sets update region - all drawing beyond this region must be ignored.
   This is used between begin_display() and end_display().
   Clip everything that is beyond the specified
   rectangle. If your system doesn't supply such a service, you
   will have hard times dealing with font cutoffs, hehe. After
   calling begin_display(), the region is assumed to clip to entire
   screen - the drawing still may occur offscreen, so clip it.
   set_region( 0, 0, 0, 0 ) is called to reset the clipping region
   to entire screen - handle this too.
*/
void SDLSystem::set_region (UINT x, UINT y, UINT width, UINT height)
{
	SDL_Rect clip;

	if (config->getVerbose())
		printf("set_region, x: %i, y:%i, width: %i, height:%i\n", x, y, width, height);

	/* Check if region is to be reset */
	if (x == 0 && y == 0 && width == 0 && height == 0)
	{
		SDL_SetClipRect(current, NULL);
		return;
	}
	clip.x = x;
	clip.y = y;
	clip.w = width;
	clip.h = height;
	
	SDL_SetClipRect(current, &clip);
}

/* Blitts a piece of an image to the drawing surface.
   Typically, (x,y) is the destination position, (i_x,i_y,i_width,i_height)
   specify the rectangle from the given image to copy. Taking advantage
   of hardware blitter is strongly recommended (unless you're dealing
   with drawing surface in system RAM...). In some ugly cases (caused
   by the author of the container) the actual rectangle may refer to
   a non-existing portion of the image (for example if image width
   is 10, x potentially can be greater than 10). You have to handle
   such ugly cases if your OS doesn't handle them.
*/
void SDLSystem::blt (IMAGE image, UINT x, UINT y, UINT i_x, UINT i_y, UINT i_width, UINT i_height)
{
	SDL_Rect srcrect;
	SDL_Rect dstrect;
	
	if (config->getVerbose())
		printf("blt, x: %i, y: %i, i_x:%i, i_y:%i, i_width: %i, i_height: %i\n", x, y, i_x, i_y, i_width, i_height);
		
	srcrect.x = i_x;
	srcrect.y = i_y;
	srcrect.w = i_width;
	srcrect.h = i_height;

	dstrect.x = x;
	dstrect.y = y;
	dstrect.w = x+i_width;
	dstrect.h = y+i_height;

	SDL_BlitSurface((SDL_Surface*)image, &srcrect, current, &dstrect);
}

/* As above, but stretching. 
   The entire image is copied and stretched to the given rectangle
   on the drawing surface. This is more difficult than blt(),
   especially taking into account special cases.
*/
void SDLSystem::stretch_blt (IMAGE image, UINT x, UINT y, UINT width, UINT height)
{
	if (config->getVerbose())
		printf("strech - x:%i, y:%i, width: %i, height: %i\n", x, y, width, height);

	SDL_StretchSurface((SDL_Surface*)image, current, x, y, width, height);	
}

/* Draws a vertical line. dwColor is 24-bit RGB (lsb is red). */
void SDLSystem::v_line (UINT x, UINT y, UINT length, DWORD color)
{
	color = SWAP24(color);

	unsigned int i		= 0;

	Uint8 red;
	Uint8 green;
	Uint8 blue;

	toComponents(color, &red, &green, &blue);

	Uint8* pixels		= (Uint8*)current->pixels;
	Uint32* pixel		= NULL;
	Uint32 lineColor 	= SDL_MapRGB(current->format, red, green, blue);
	Uint32 startX		= current->clip_rect.x;
	Uint32 endX		= current->clip_rect.x+current->clip_rect.w-1;
	Uint32 startY		= current->clip_rect.y;
	Uint32 endY		= current->clip_rect.y+current->clip_rect.h-1;

	/* First, check if the line is completly outside the clipping area */
	if (x < startX || x > endX)
		return;
	if (y+length-1 < startY || y > endY)
		return;
		
	/* Clip the x coordinates if nescessary */
	if (y < startY)
	{
		length -= startY-y;
		y = startY;
	}
	if (y+length-1 > endY)
		length = endY-y+1;

	
	pixels 	+= y * current->pitch + (x * 4);
	
	if (config->getVerbose())
		printf("v_line, x: %i y: %i, length: %i\n", x, y, length);
	
	for (i = 0; i < length; i++)
	{
		pixel = (Uint32*)pixels;
		*pixel = lineColor;
		pixels += current->pitch;
	}
}

/* Draws a horizontal line.*/
void SDLSystem::h_line (UINT x, UINT y, UINT length, DWORD color)
{
	color = SWAP24(color);

	unsigned int i		= 0;

	Uint8 red;
	Uint8 green;
	Uint8 blue;

	toComponents(color, &red, &green, &blue);

	Uint8* pixels		= (Uint8*)current->pixels;
	Uint32* pixel		= NULL;
	Uint32 lineColor 	= SDL_MapRGB(current->format, red, green, blue);
	Uint32 startX		= current->clip_rect.x;
	Uint32 endX		= current->clip_rect.x+current->clip_rect.w-1;
	Uint32 startY		= current->clip_rect.y;
	Uint32 endY		= current->clip_rect.y+current->clip_rect.h-1;
	
	if (config->getVerbose())
		printf("h_line, x: %i y: %i, length: %i\n", x, y, length);

	/* First, check if the line is completly outside the clipping area */
	if (y < startY || y > endY)
		return;
	if (x+length-1 < startX || x > endX)
		return;
		
	/* Clip the x coordinates if nescessary */
	if (x < startX)
	{
		length -= startX-x;
		x = startX;
	}
	if (x+length-1 > endX)
		length = endX-x+1;
	
	pixel = (Uint32*)(pixels + y * current->pitch + (x * 4));

	for (i = 0; i < length; i++)
	{
		*pixel = lineColor;
		pixel++;
	}
}

void SDLSystem::toComponents(DWORD color, Uint8* red, Uint8* green, Uint8* blue)
{
	*red 		= color>>16;
	*green 		= color>>8;
	*blue 		= color;
}

/* Draws a solid-filled rectangle.*/
void SDLSystem::fill_rect (UINT x, UINT y, UINT width, UINT height, DWORD color)
{
	color = SWAP24(color);

	if (config->getVerbose())
		printf("fill_rect - x: %i, y: %i, width: %i, height: %i, colour: %x\n", x, y, width, height, color);
	
	Uint8 red;
	Uint8 green;
	Uint8 blue;

	toComponents(color, &red, &green, &blue);

	SDL_Rect rect;
	rect.x = x;
	rect.y = y;
	rect.w = width;
	rect.h = height;
	
	SDL_FillRect(current, &rect, SDL_MapRGB(current->format, red, green, blue));
}

/* Displays print dialog boxes and prints selected pages using
Panorama::print_callback(). The callback uses stretch_blt(),
v_line, h_line() and out_text() to draw particular objects.
The SI must distinguish calls to these methods during printing
from normal drawing calls.*/
void SDLSystem::print( const char* jobname)
{
//	printf("print\n");
	
}

/* Makes the cursor visible or hidden. The cursor will not be drawn if the
image wasn't set. On startup the status should be: hidden.*/
void SDLSystem::show_cursor (bool show)
{
	if (config->getVerbose())
		printf("show_cursor\n");
	
	cursorShow = show;
}

/* Moves the cursor to another location.
   If you are making a SI for a primitive OS, like DOS, you will be happy
   that throughout the core sources a single-threaded approach is assumed.
   In case of mouse cursor movement, you simply supply the notification
   through get_message() posting MSG_MOUSEMOVE or MSG_MOUSEDRAG. Then
   Panorama handling the message queue itself calls move_cursor() for
   you to paint the cursor at the right place on the screen.

   However if your system isn't so primitive as DOS is, you will want to
   implement cursor movement on another thread, supplying a low-latency
   cursor display. To do this, make your own cursor routine and call it
   from your mouse thread, and leave move_cursor() empty. You still have
   to post MSG_MOUSEMOVE and MSG_MOUSEDRAG to Panorama to indicate mouse
   movement and thus allow handling links, etc., but the move_cursor()
   will no longer paint the cursor. I perform like this in the BeOS SI
   since version 2.0.12 - don't hesitate to see how it is done (mouse.h
   and mouse.cpp). I've tried to do like this in the Win32 SI, too, but
   Win32 itself seems to be not enough preemptive...

   Note that in windowed mode you should be careful when the mouse gets
   out of the window. You can handle this as a special case, and e.g.
   hide the cursor (like in the BeOS SI) or simply leave the cursor
   visible where the mouse left the window (like in the Win32 SI, where
   for shit I can't figure out when the cursor leaves the window).
*/
void SDLSystem::move_cursor (UINT x, UINT y)
{
	if (cursorShow)
	{
		SDL_Rect srcrect;
		SDL_Rect dstrect;
		
		srcrect.x = 0;
		srcrect.y = 0;
		srcrect.w = cursorImage->w;
		srcrect.h = cursorImage->h;
		
		dstrect.x = x-cursorX;
		dstrect.y = y-cursorY;

		//Delete old pointer
		SDL_BlitSurface(current, &cursorArea, videoSurface, &cursorArea);
		SDL_BlitSurface(cursorImage, &srcrect, videoSurface, &dstrect); 
		SDL_Flip(videoSurface);
		
		cursorArea.x = x-cursorX;
		cursorArea.y = y-cursorY;
		cursorArea.w = cursorImage->w;
		cursorArea.h = cursorImage->h;
		
	}
}

/* Sets new cursor image.
   Sets another cursor image. Take care of the hotspot !
*/
void SDLSystem::set_cursor (IMAGE image, UINT xspot, UINT yspot)
{
	UINT x = cursorArea.x+cursorX;
	UINT y = cursorArea.y+cursorY;
	cursorImage 	= (SDL_Surface*) image;
	cursorX 	= xspot;
	cursorY 	= yspot;
	move_cursor(x, y);
}

/* Loads a sound from a buffer.
   Loads a sound. "buf" is the sound data. "size" is in bytes.
   "flags" indicate whether the sound is 16-bit or stereo (interleaved).
   8-bit sounds are unsigned while 16-bit are signed - this is exactly
   like in the WAV file. SOUND is another mysterious (void*) pointer of
   your choice - cast it.

   enum _SOUND_FLAGS {
   SND_STEREO = 1,
   SND_16BIT = 2

*/
SOUND SDLSystem::load_sound (BUFFERPTR buf, DWORD size, DWORD freq, UINT flags)
{
	if (config->getVerbose())
		printf("load_sound, buf: %x, size: %i, freq: %i, flags: %x\n", buf, size, freq, flags);
	
	if (config->getSoundEffects())
	{
		SDL_AudioCVT cvt;
		Uint16 srcFormat;
		Uint8 channels;
		
		if (flags & SND_16BIT)
			srcFormat = AUDIO_S16SYS;
		else
			srcFormat = AUDIO_U8;
		
		if (flags & SND_STEREO)
			channels = 2;
		else
			channels = 1;

		/* Convert all samples to the right format */
		if (SDL_BuildAudioCVT(&cvt, srcFormat, channels, freq, AUDIO_S16SYS, 2, config->getMixingFrequency() ) == -1)
		{
			printf("Sample conversion not possible\n");
			return NULL;
		}
		cvt.buf =  (Uint8*) malloc( size * cvt.len_mult );
		
		cvt.len = size;
		memcpy(cvt.buf, buf, size);
	
		SDL_ConvertAudio(&cvt);
		Sample* sample = new Sample(cvt.buf, (Uint32) (cvt.len * cvt.len_ratio) );
		
		return sample;
	}
	else
	{
		return NULL;
	}
}

/* Plays a loaded sound. 
   Plays a sound. May be called while the sound is already playing -
   restart it in this case.
*/
void SDLSystem::play_sound (SOUND sound)
{
	if (config->getVerbose())
		printf("play_sound : %x\n", sound);
	
	if (config->getSoundEffects())
		soundSystem->play((Sample*) sound);
}

/* Removes a sound from memory. 
   Obvious. Sound management done within core sources, just like image
   management.*/
void SDLSystem::unload_sound (SOUND sound)
{
	if (config->getVerbose())
		printf("unload_sound : %x\n", sound);
	
	if (config->getSoundEffects())
	{
		Sample* sample = (Sample*) sound;
		delete sample;	
	}
}

/* Play a song stored in memory 
   Plays a song from memory. The type is indicated by filename extension.
*/
bool SDLSystem::play_song (const char* filename, bool loop, BUFFERPTR buf, DWORD size)
{
	if (config->getVerbose())
		printf("play_song from memory: %s, loop: %i, pointer: %x, size: %i\n", filename, loop, buf, size);

	if (config->getMusic())
                return soundSystem->play(filename, buf, size, loop);
	else
		return false;
}

/* Play a song stored in current directory 
   Plays a song from a file. Argument can be either full path specified
   by you in get_user_string() or file name/short path specified by
   the author of the container.
*/
bool SDLSystem::play_song (const char* filename, bool loop)
{
	if (config->getVerbose())
		printf("play_song from file: %s, loop: %i\n", filename, loop);
	
	if (config->getMusic())
	{
		char* suffix = (char*)filename+strlen(filename)-4;

		if (strcasecmp(suffix, ".mo3") == 0)
		{
			char* unpacked = platform->unmo3((char*)filename);
			if (unpacked != NULL)
				filename = unpacked;
			else
				return false;
				
		}
		
                return soundSystem->play(filename, loop);
	}
	else
		return false;
}

/* Stop (and unload from memory) a song 
   Stops a song. If b_fade_out is true, you should first slowly downgrade
   the volume before stopping the song.
*/
void SDLSystem::stop_song (bool b_fade_out)
{
	if (config->getVerbose())
		printf("stop_song, fade out: %i\n", b_fade_out);
	
	if (config->getMusic())
		soundSystem->stop(b_fade_out);
}

/* Pause/unpause a loaded song 
   Hmmm... Actually not used, but the meaning is obvious.
*/
void SDLSystem::pause_song()
{
	if (config->getVerbose())
		printf("pause_song\n");

	if (config->getMusic())
		soundSystem->pause();	
}

/* Control overall sound volume (>0 - louder, <0 - quieter) 
   How much ? Your choice.
*/
void SDLSystem::volume_control (int direction)
{
	if (config->getVerbose())
		printf("volume_control: %i\n", direction);
	
	if (config->getMusic())
	{
		if (direction < 0)
			soundSystem->decreaseVolume();
		else
			soundSystem->increaseVolume();
	}
}

/* Prepares a movie for playback. Also returns framerate (fps)*/
MOVIE SDLSystem::load_movie (const char* filename, bool loop, UINT* fps)
{
//	printf("load_movie\n");
	
	return NULL;
}

/* Advances movie frame */
void SDLSystem::update_frame (MOVIE movie)
{
//	printf("update_frame\n");
	
}

/* Displays a movie frame */
void SDLSystem::blt (MOVIE movie, UINT x, UINT y, UINT width, UINT height)
{
//	printf("blt movie\n");
	
}

/* Unloads the movie */
void SDLSystem::unload_movie (MOVIE movie)
{
//	printf("unload_movie\n");
	
}

/* Gets current system time in millisec. */
DWORD SDLSystem::get_time()
{
	return SDL_GetTicks();
}

/* Starts a timer (issuing a message every "period" ms).
   Starts a timer. The specified message should be posted at a rate
   specified by "period" (in milliseconds).
*/
void SDLSystem::start_timer (UINT msg, UINT period)
{
	SDLTimer* timer = NULL;
	if (config->getVerbose())
		printf("start_timer, msg: %i, periode %i\n", msg, period);

	timer = (SDLTimer*) malloc ( sizeof(SDLTimer) );
	timer->message 	= msg;
	timer->period 	= period;
	timer->timerID 	= SDL_AddTimer(period, SDLTimer_callback, timer);
	timer->system 	= this;
	SDLTimer_addTimer( timer );
}

/* Stops a timer.
   Stops a timer. Do not assume that this is always called (i.e. release
   any timers when the program shuts down).
*/
void SDLSystem::stop_timer (UINT msg)
{
	SDLTimer* timer = NULL;
	if (config->getVerbose())
		printf("stop_timer, msg: %i\n", msg);
	timer = SDLTimer_removeTimer( msg );
	if (timer)
	{
		SDL_RemoveTimer(timer->timerID);
		free(timer);
	}
}

/* Retrieves a message (if any).
   If your OS doesn't implement message queue, you have to do it
   on your own. This function gets a message from the queue.
   If no message is in the queue, it returns false. The exact
   list of messages is in the system.h. Convert any OS-specific
   messages to these Panorama messages.
*/
bool SDLSystem::get_message (UINT& msg, DWORD& param)
{
	SDL_Event event;
	Uint32 x = 0;
	Uint32 y = 0;
	Uint8 buttonState = 0;
	Message* message = NULL;
	static bool mouseLeftDown = false; //Part of a hack, see description where used.

		

	// First check the internal message queue 
	SDL_mutexP(postSync);
	message = (Message*) Queue_dequeue(messageQueue);
	Message* next = (Message*)Queue_peep(messageQueue);
	while (message != NULL && next != NULL && next->msg == message->msg && next->param == message->param)
	{
		Queue_dequeue(messageQueue);
                free(next);
		next = (Message*)Queue_peep(messageQueue);
	}
	SDL_mutexV(postSync);

	if (message != NULL)
	{
		msg 	= message->msg;
		param 	= message->param;
		free(message);
	
		return true;
	}
	


	if ( SDL_PollEvent(&event) )
	{
		switch (event.type)
		{
			SDLKey sym; 
			SDLMod mod;
			case SDL_KEYDOWN:
				sym = event.key.keysym.sym; 
				mod = event.key.keysym.mod;
				msg = MSG_NOTHING;
				if (sym == SDLK_ESCAPE || sym == SDLK_BACKSPACE)
				{
					msg 	= MSG_RETURN;
					param 	= 1;
				}
				else if (sym == SDLK_F10 || ( (sym == SDLK_q || sym == SDLK_x) && (KMOD_ALT & mod) ))
					msg = MSG_QUIT;
				else if (sym == SDLK_F1)
					msg = MSG_HELP;
				else if (sym == SDLK_F7)
					msg = MSG_SEARCH;
				else if (sym == SDLK_F2)
					msg = MSG_SAVE;
				else if (sym == SDLK_F3)
				{
					msg = MSG_PLAY;
					param = PLAY_SLOT_0;
				}
				else if (sym == SDLK_F4)
				{
					msg = MSG_TOGGLEEFFECTS;
					config->setEffects( !config->getEffects());
				}
				else if (sym == SDLK_F12)
					msg = MSG_RELOAD;
				else if (sym == SDLK_HOME || (sym == SDLK_PAGEUP && KMOD_CTRL & mod))
				{
					msg = MSG_SCROLL;
					param = SCROLL_HOME;
				}
				else if (sym == SDLK_END || (sym == SDLK_PAGEDOWN && KMOD_CTRL & mod))
				{
					msg = MSG_SCROLL;
					param = SCROLL_END;
				}
				else if (sym == SDLK_PAGEUP || (sym == SDLK_UP && KMOD_SHIFT & mod) )
				{
					msg = MSG_SCROLL;
					param = SCROLL_PAGEUP;
				}
				else if (sym == SDLK_PAGEDOWN || (sym == SDLK_DOWN && KMOD_SHIFT & mod))
				{
					msg = MSG_SCROLL;
					param = SCROLL_PAGEDOWN;
				}
				else if (sym == SDLK_UP)
				{
					msg = MSG_SCROLL | MSG_PREVLINK;
					param = SCROLL_UP;
				}
				else if (sym == SDLK_DOWN)
				{
					msg = MSG_SCROLL | MSG_NEXTLINK;
					param = SCROLL_DOWN ;
				}
				else if (sym == SDLK_LEFT && KMOD_CTRL & mod)
				{
					msg 	= MSG_RETURN;
					param 	= 0;
				}
				else if (sym == SDLK_LEFT)
				{
					msg = MSG_SCROLL;
					param = SCROLL_LEFT;
				}
				else if (sym == SDLK_RIGHT && KMOD_CTRL & mod)
					msg = MSG_ACTIVATELINK;
				else if (sym == SDLK_RIGHT)
				{
					msg = MSG_SCROLL;
					param = SCROLL_RIGHT;
				}
				else if (sym == SDLK_SPACE)
				{
					msg = MSG_SCROLL;
					param = SCROLL_PRECISE;
				}
				else if (sym == SDLK_RETURN && (KMOD_CTRL & mod))
					msg = MSG_COPYLINK;
				else if (sym == SDLK_RETURN && (KMOD_LALT & mod))
					SDL_WM_ToggleFullScreen(videoSurface);
				else if (sym == SDLK_RETURN)
					msg = MSG_ACTIVATELINK;
				else if (sym == SDLK_1 && (KMOD_CTRL & mod))
				{
					msg = MSG_BACKGROUND;
					param = 1;
				}
				else if (sym == SDLK_2 && (KMOD_CTRL & mod))
				{
					msg = MSG_BACKGROUND;
					param = 2;
				}
				else if (sym == SDLK_3 && (KMOD_CTRL & mod))
				{
					msg = MSG_BACKGROUND;
					param = 3;
				}
				else if (sym == SDLK_4 && (KMOD_CTRL & mod))
				{
					msg = MSG_BACKGROUND;
					param = 4;
				}
				else if (sym == SDLK_5 && (KMOD_CTRL & mod))
				{
					msg = MSG_BACKGROUND;
					param = 5;
				}
				else if (sym == SDLK_6 && (KMOD_CTRL & mod))
				{
					msg = MSG_BACKGROUND;
					param = 6;
				}
				else if (sym == SDLK_7 && (KMOD_CTRL & mod))
				{
					msg = MSG_BACKGROUND;
					param = 7;
				}
				else if (sym == SDLK_8 && (KMOD_CTRL & mod))
				{
					msg = MSG_BACKGROUND;
					param = 8;
				}
				else if (sym == SDLK_9 && (KMOD_CTRL & mod))
				{
					msg = MSG_BACKGROUND;
					param = 9;
				}
				else if (sym == SDLK_1)
				{
					msg = MSG_PLAY;
					param = PLAY_SLOT_1;
				}
				else if (sym == SDLK_2)
				{
					msg = MSG_PLAY;
					param = PLAY_SLOT_2;
				}
				else if (sym == SDLK_3)
				{
					msg = MSG_PLAY;
					param = PLAY_SLOT_3;
				}
				else if (sym == SDLK_4)
				{
					msg = MSG_PLAY;
					param = PLAY_SLOT_4;
				}
				else if (sym == SDLK_5)
				{
					msg = MSG_PLAY;
					param = PLAY_SLOT_5;
				}
				else if (sym == SDLK_6)
				{
					msg = MSG_PLAY;
					param = PLAY_SLOT_6;
				}
				else if (sym == SDLK_7)
				{
					msg = MSG_PLAY;
					param = PLAY_SLOT_7;
				}
				else if (sym == SDLK_8)
				{
					msg = MSG_PLAY;
					param = PLAY_SLOT_8;
				}
				else if (sym == SDLK_0)
				{
					msg = MSG_STOP;
				}
				else if (sym == SDLK_PLUS)
				{
					msg = MSG_SETVOLUME;
					param = 1;
				}
				else if (sym == SDLK_MINUS)
				{
					msg = MSG_SETVOLUME;
					param = 0-1;
				}
				else if (sym == SDLK_c)
				{
					msg = MSG_SETPLAYMODE;
					param = PLAY_CYCLE;
				}
				else if (sym == SDLK_r)
				{
					msg = MSG_SETPLAYMODE;
					param = PLAY_RANDOM;
				}
				else if (sym == SDLK_l)
				{
					msg = MSG_SETPLAYMODE;
					param = PLAY_LOOP;
				}
				else if (sym == SDLK_p && (KMOD_CTRL & mod))
					msg = MSG_PRINT;

				if (msg == MSG_NOTHING)
					return false;
				return true;				
			case SDL_MOUSEMOTION:

                                buttonState = SDL_GetMouseState((int*)&x, (int*)&y);
				param = (event.motion.x<<16|event.motion.y);

                                if ( SDL_BUTTON(buttonState) & SDL_BUTTON_LEFT)
                                        msg = MSG_MOUSEDRAG;
                                else
                                {
                                        /* This is a hack! In windowed mode, SDL generates a MOUSEMOTION event between MOUSEDOWN
                                        and MOUSEUP even though the mouse does not move - to get around this we convert to a MOUSEDRAG
                                        since panorama functions fine with that.*/
                                        if (!mouseLeftDown)
                                                msg = MSG_MOUSEMOVE;
                                        else
                                                msg = MSG_MOUSEDRAG;
                                }


				/* Remove any following mouse motion events. */
				while (SDL_PeepEvents(&event, 1, SDL_PEEKEVENT, 0xffffffff) == 1)
				{
					
					if (event.type == SDL_MOUSEMOTION)	
					{
						SDL_PollEvent(&event);
						param = (event.motion.x<<16|event.motion.y);
					}
					else
					{
						break;
					}	
				}

				return true;	

			case SDL_MOUSEBUTTONDOWN:
				x = event.button.x;
				y = event.button.y;
				if (event.button.button == SDL_BUTTON_LEFT)
				{
					mouseLeftDown = true;
					msg = MSG_MOUSELBDOWN;
					param = (x<<16 | y);  
				}
				else if (event.button.button == SDL_BUTTON_RIGHT)
				{
					msg = MSG_MOUSERBDOWN;
					param = (x<<16 | y);  
				}
				else if(event.button.button == SDL_BUTTON_WHEELUP)
				{
					msg = MSG_MOUSEWHEEL;
					param = (DWORD)1;
				}
				else if(event.button.button == SDL_BUTTON_WHEELDOWN)
				{
					msg = MSG_MOUSEWHEEL;
					param = (DWORD)-1;
				}
				else
				{
					break;
				}
				return true;
			case SDL_MOUSEBUTTONUP:
				if (event.button.button == SDL_BUTTON_LEFT)
				{
					mouseLeftDown = false;
					msg = MSG_MOUSELBUP;
				}
				else if (event.button.button == SDL_BUTTON_RIGHT)
					msg = MSG_MOUSERBUP;
				else
					break;
				param = (event.button.x<<16 | event.button.y);  
				return true;
		}
	}
	msg = MSG_NOTHING;
	return false;
	
}

/* Calls idle processing routine.*/
void SDLSystem::idle()
{
	/* Just sit tight and wait for an event, when one comes - push it back in the queue so it can be 
	handled */
	SDL_Event event;
	SDL_WaitEvent(&event);
	SDL_PushEvent(&event);
	
}

/* Posts a message. */
void SDLSystem::post_message (UINT msg, DWORD param)
{
        /* Create message */
	Message* message 	= (Message*) malloc( sizeof(Message) );
	message->msg 		= msg;
	message->param 		= param;

        /* Added it to custom message queue. */
	SDL_mutexP(postSync);
	Queue_enqueue(messageQueue, message);
	SDL_mutexV(postSync);
	

        /* Because the SDL event thread could be sleeping, the system may not automatically pick up on the posted messages - so we
        add a dummy event to the SDL event queue to force it to wakeup, and thus find this message. */
        if (msg == MSG_NEXTIMAGE)
        {
                int x;
                int y;
                Uint8 mouse =  SDL_GetMouseState(&x, &y);
                SDL_Event event;
                event.type = SDL_MOUSEMOTION;
                event.motion.state = mouse;
                event.motion.x = x;
                event.motion.y = y;
                event.motion.xrel = 0;
                event.motion.yrel = 0;
                SDL_PushEvent(&event);
        }
}

/* Gets user flag
   This function is called when the engine starts up. You have to respond
   to a flag request.
      CFG_LOGPIXELSY - return 96. This should be a typical setting.
		       I don't know if it will be used in the future.
      CFG_EFFECTS - the user may choose to disable the effects at startup.
		    true if enabled, false if disabled.
      CFG_FIRSTTUNE - if there is an external tune, return 0. Otherwise
		      return 1.
      CFG_IMAGE_ALIGN - return 1, 2, 4 or 8 (or 16?), depending on what
			boundary you want the physical width of images
			to be aligned. Note that in create_image()
			you receive the "buf"'s real width aligned to this
			boundary, but the "width" parameter specifies
			an unaligned, logical width.
      CFG_IMAGE_BYTES_PER_PIXEL - return 3 or 4, depending on how many bytes
				  per pixels you want to receive in
				  create_image().
      CFG_IMAGE_REVERSE_RGB - true if you want to get BGR (or BGRA) byte
			      order images. Otherwise return false - you
			      will get RGB (or RGBA) order. If you are not
			      sure, experiment with this setting and see
			      what happens.
      CFG_REVERSE_RGB - if true, all colors will have reversed components
			(blue will come in bits 7..0). Otherwise the
			order will be normal (red on bit 7..0, blue on
			bits 23..16).
      CFG_FAST_JPEG - true if the user wants to use fast integer IDCT
		      instead of the default floating point IDCT. As far
		      as I can see the two methods don't differ much on
		      my Athlon, at least when compiled by MSVC.
*/
UINT SDLSystem::get_user_flag (UINT flag)
{
	if (config->getVerbose())
		printf("get_user_flag, flag: %i\n", flag);
	
	switch (flag)
	{
		case CFG_LOGPIXELSY:
			return 96;
		case CFG_EFFECTS:
			return config->getEffects();
		case CFG_FIRSTTUNE:
			if (config->getExternalTune() != NULL)
				return 0;
			else
				return 1;
		case CFG_IMAGE_ALIGN:
			return 1;
		case CFG_IMAGE_BYTES_PER_PIXEL:
			return 4;
		case CFG_IMAGE_REVERSE_RGB:
			return true;
		case CFG_REVERSE_RGB:
#if __BYTE_ORDER == __BIG_ENDIAN
	            return true;
#else
                  return false;
#endif
		case CFG_FAST_JPEG:
			return false;
		default:
			fprintf(stderr, "Invalid flag given in get_user_flag. Returning 0 and hopeing for the best.");
	}
	return 0;
}

/* Gets user string
   Here are only two settings: CFG_OVERRIDE_ARTICLE and CFG_SONG.
   Both can be entered at commandline (override article only in
   editor's version). The string is duplicated right after this
   function exits.
*/

char* SDLSystem::get_user_string (UINT flag)
{
	if (config->getVerbose())
		printf("get_user_string\n");

	switch (flag)
	{
		case CFG_OVERRIDE_ARTICLE:
			return config->getArticle();
		case CFG_SONG:
			return config->getExternalTune();
		default:
			fprintf(stderr, "Invalid flag given in get_user_string. Returning NULL and hopeing for the best.");
	}
	return NULL;
}

/* Opens http:// mailto: ftp: or anything else*/
void SDLSystem::open_link (const char* text)
{
	if ( config->getVerbose() )
		printf("open_link: %s\n", text);
	
	platform->handleExternalLink( (char*) text, config->getExternalLink() );	
}

/* Puts a string into clipboard.*/
void SDLSystem::put_to_clipboard (const char* text)
{
	if ( config->getVerbose() )
		printf("put_to_clipboard: %s\n", text);
		
	platform->toClipboard( (char*)text );
	
}

bool isScanning = false;
Uint32 articleNo = 0;
/* Building search database*/
bool SDLSystem::scan (UINT progress)
{
	if (isScanning == false)
	{
		articleNo 	= 0;
		isScanning 	= true;
		panorama.scan_callback();
	}
	else
	{
		articleNo++;
	}
	return true;
}

char* SDLSystem::extractWord(Vector* word)
{
	char* string = (char*) malloc( Vector_getSize(word)+1 );
	
	for (int i = 0; i < Vector_getSize(word); i++)
	{
		const void* next = Vector_get(word, i);
		string[i] = *reinterpret_cast<const char*>( &next );
	}
	string[Vector_getSize(word)] = 0;
	return string;
}

void SDLSystem::findWords(SearchDatabaseTree* tree, Vector* word, Vector* wordCollection)
{
	assert (tree != NULL);
	
	/* First handle all letters that comes before this one in the alphabet */
	if (tree->get_less() != NULL)
		findWords(tree->get_less(), word, wordCollection);

	/* Now, add this letter to the word we are buildning */
	int letter = tree->get_letter();
	Vector_add(word, reinterpret_cast<void*>(letter));
	
	/* If this combination is a complete word, extract and save it */
	if (tree->is_end_letter())
	{
		char* wordString = extractWord(word);
		Vector_add(wordCollection, wordString);
	}
	
	/* Try to add more letters to this word */
	if (tree->get_equal() != NULL)
		findWords(tree->get_equal(), word, wordCollection);

	/* Remove this letter, and attempt to construct words with letters later in the alphabet */
	Vector_remove(word, Vector_getSize(word)-1 );
	
	if (tree->get_greater() != NULL)
		findWords(tree->get_greater(), word, wordCollection);
}

SearchDatabaseTree* SDLSystem::find(SearchDatabaseTree* sdb, char* word)
{
	Uint32 i = 0;

	while (word[i] != 0)
	{
		unsigned char letter = word[i];

		if (sdb->get_letter() == letter && word[i+1] == 0)
		{
			break;
		}
		else if (sdb->get_letter() == letter)
		{
			sdb = sdb->get_equal();
			if (sdb == NULL)
				return NULL;
			i++;
		}
		else if (sdb->get_letter() < letter)
		{
			sdb = sdb->get_greater();
			if (sdb == NULL)
				return NULL;
		}
		else if (sdb->get_letter() > letter)
		{
			sdb = sdb->get_less();
			if (sdb == NULL)
				return NULL;
		}
	}
	return sdb;
}

void SDLSystem::findAll(SearchDatabaseTree* sdb, char* beginsWith, Vector* words)
{
	for(Uint32 i = 0; beginsWith[i] != 0; i++)
		beginsWith[i] = tolower(beginsWith[i]);
	
	Vector* characterQueue = Vector_construct();
	
	SearchDatabaseTree* word = NULL;
	
	/* Special case, no start string - just grab all words */
	if (strlen(beginsWith) == 0)
	{
		findWords(sdb, characterQueue, words);
		return;
	}

	/* Find the node that represents the last letter of the beginsWith string */
	word = find(sdb, beginsWith);
	
	if (word == NULL)	// No string starts with beginsWith
		return;
	
	if (word->is_end_letter())	// beginsWith is a word in itself
		Vector_add(words, beginsWith);
	
	/* Go find any extensions of the word */
	word = word->get_equal();
	if (word == NULL)	//There are no extensions of the word
		return;
	
	for (Uint32 i = 0; beginsWith[i] != 0; i++)
	{
		int next = (int) beginsWith[i];
		Vector_add(characterQueue, reinterpret_cast<void*>(next));
	}
	
	findWords(word, characterQueue, words);
	
	Vector_shallowFree(characterQueue);
}

void SDLSystem::findArticles(SearchDatabase* database, char* word, Vector* articles)
{
	SearchDatabaseTree* wordNode = find(database->get_database_tree(), word);
	
	if (wordNode != NULL)
	{
		unsigned char* set = wordNode->get_article_set();
		for (Uint32 i = 0; i < database->get_num_articles(); i++)
		{
			if ( (set[i>>3] & (1<<(i&7)) ) != 0 ) //This line has been copied from the beos version. Would be nice with at description of that format set? Is it perhaps a bitfield marking wheter an article is included?
			{
				char* articleName = database->get_article_name(i);
				char* title = database->get_article_title(i);

				if (title == NULL)
					title = articleName;
				
				Uint32 length = 0;
				if (articleName != NULL)
					length += strlen(articleName);
				if (title != NULL)
					length += strlen(title);
				
				length += strlen(" - ");
				
				char* articleString = (char*) malloc( length + 1);

				sprintf(articleString, "%s%c%s", title, 0, articleName);
								
				Vector_add(articles, articleString);
			}
		}
	}
}

/* Article search dialog */
bool SDLSystem::search (SearchDatabase* sdb, char* &wordret, char* &article)
{
	if (config->getVerbose())
		printf("search\n");
	
	/* Lock the message queue, so it does not fill up while we are doing this */
	SDL_mutexP(postSync);

	SearchDatabaseTree* tree = sdb->get_database_tree();
	
	/* Display search GUI */
	bool cont 	= true;
	SDL_Event event;

	findAll(tree, "", searchWords);
	searchPanel->resetList();
	
	searchPanel->show();
	while (cont)
	{
		searchPanel->paint(videoSurface, 0, 0);
		SDL_Flip(videoSurface);

		if( SDL_PollEvent(&event))
			searchPanel->handleEvent(&event, 0, 0);

		if (searchPanel->textChanged())
		{
			/* Release all memory allocated to prior search strings */
			for (int i = 0; i < Vector_getSize(searchWords); i++)
				free(Vector_get(searchWords, i));
			
			Vector_empty(searchWords); //empty the vector
			
			findAll(tree, searchPanel->getText(), searchWords);
			searchPanel->resetList();
		}
		else if (searchPanel->wordSelectionChanged())
		{
			char* word = searchPanel->getSelectedWord();
			if (word != NULL)
			{
				/* Release all memory allocated to prior article strings */
				for (int i = 0; i < Vector_getSize(searchArticles); i++)
					free(Vector_get(searchArticles, i));

				Vector_empty(searchArticles);
				findArticles(sdb, word, searchArticles);
			}
		}
		cont = !searchPanel->done();
	}

	char* searchWord 	= searchPanel->getWord();
	char* searchArticle 	= searchPanel->getArticle();
	
	/* Erase the search panel */
	SDL_Rect rect;
	searchPanel->getPosition(&rect);
	
	SDL_BlitSurface(current, &rect, videoSurface, &rect);
	SDL_Flip(videoSurface);
	
	/* Clear the contents of the panel */
	Vector_empty(searchArticles);
	searchPanel->hide();

	
	SDL_mutexV(postSync);

	if( searchWord != NULL && searchArticle != NULL)
	{
		/* Find the place in the searchArticle where the article file name starts */
		searchArticle = searchArticle+strlen(searchArticle) + 1;
		
		/* We have to make copies of the return values, since the core system re-uses the string,
		which make it unsafe for me to free the memory again - this is an actual memory leak (eek)
		but you will have to search ALOT of times for it to be of any significance */
		char* copyWord 		= (char*) new char[ strlen(searchWord) + 1];
		char* copyArticle 	= (char*) new char[ strlen(searchArticle) + 1];
		sprintf(copyWord, "%s", searchWord);
		sprintf(copyArticle, "%s", searchArticle);
		wordret = copyWord;
		article = copyArticle;
		
		return true;
	}
	
	return false;
}


/* Checks a pointer if non-NULL. If NULL - makes the system shut down.*/
void SDLSystem::critical_check (void* pointer)
{
	if (config->getVerbose())
		printf("critical_check, %x\n", pointer);
	
	if (pointer == NULL)
	{
		fprintf(stderr, "critical_check failed! Shutting down ....\n");
		exit(1);
	}
}

/* Sets an error message displayed upon exit. NULL if no message to be d.
   The error message (can be NULL) set by this function should
   appear after the program quits. The message is a constant string,
   so you can count on it being always at the same place - you don't
   have to duplicate it, but also don't modify it.
*/
void SDLSystem::set_error_message (char* message)
{
	if (config->getVerbose())
		printf("set_error_message : %s\n", message);
		
	errorMessage = message;
}

/* Output debug message (used only in debug versions) 
   Legacy. Leave that empty, unless you want to use it for some purpose.
*/
void SDLSystem::debug_message (const char* message)
{}

char* SDLSystem::getErrorMessage()
{
	return errorMessage;
}

