#ifndef SDL_SYSTEM_H
#define SDL_SYSTEM_H

#include <stdlib.h>
#undef NULL

#include "../../core/basedefs.h"
#include "../../core/system.h"
#include "../../core/panorama.h"
#include "../../core/search.h"
#include <stdio.h>
#include "SDL/SDL.h"
#include "Vector.h"
#include "Queue.h"
#include "Map.h"
#include "SearchPanel.h"
#include "SoundSystem_dumbAndOgg.hpp"
#include "Platform.h"
#include "SDLConfig.h"
#include "FontMapping.h"


using namespace SDLGUI;

class SDLSystem : public System
{
	public:

	SDLSystem(Platform* _platform, int argc, char** argv);
	
	/* Returns TRUE if the object is available (esp. graphics)*/
	bool is_ready();
	
	/* Shuts down the system (graphics, sound, optional message) */
	~SDLSystem();
	
	/* On first call, the filename points to a mask (e.g. *.*). The
	function is called until it returns FALSE, what indicates that it
	cannot find any matching files.*/
	bool find_files (char* &filename);
	
	/* Checks if a file exists */
	bool find_one_file (char* &filename);
	
	/* Creates a file. Returns FALSE if cannot create. The handle to
	the file is kept internally in the class, as only one file
	is opened at a time. The "where" parameter specifies where
	the file is to be created. It also has a flag telling whether
	to overwrite an existing file.*/
	bool create_file (const char* filename, UINT where);
	
	/* Opens a file. Returns FALSE if cannot open. Handle as above.*/
	bool open_file (const char* filename);
	
	/* Returns size of an opened file */
	DWORD get_file_size();
	
	/* Sets file position. Returns FALSE if the pointer exceedes file size
	or if cannot set the position. */
	bool seek_file (DWORD position);
	
	/* Reads opened file.*/
	bool read_file (BUFFERPTR buf, DWORD size);
	
	/* Writes to opened file. */
	bool write_file (BUFFERPTR buf, DWORD size);
	
	/* Closes opened file.*/
	void close_file();
	
	/* Initializes window and video mode. */
	bool init_display (UINT width, UINT height);
	
	/* Sets title of the program window */
	void set_window_title (const char* title);
	
	/* Installs truetype font ("where" is the same as for create_file())*/
	void install_font (const char* filename, UINT from_where);
	
	/* Creates a font object. See the corresponding stuff in system.cpp*/
	FONT create_font (const char* face, UINT required_height, UINT flags, UINT charset, UINT& height, UINT& add_height);
	
	/* Creates an image. pBuf points to an image in 24-bit format. The system
	will probably want to convert it to a more suitable format matching
	the screen mode. pMask is either NULL or specifies a 1-bit image mask
	/ (bit=1 - transparent, bit=0 - opaque).*/
	IMAGE create_image (UINT width, UINT height, BUFFERPTR buf, BUFFERPTR mask);
	
	/* Destroys created image.*/
	void unload_image (IMAGE image);
	
	/* Returns width of a text taking into account the font.*/
	UINT get_text_width (const char* text, FONT font);
	
	/* Indicates start of drawing. This method will probably lock the
	drawing surface (maybe backbuffer). All the following drawing
	routines (except Flip()) should be called between BeginDisplay()
	and EndDisplay() - a check is appropriate. Note that also
	blt( MOVIE ) routine can be called only between Begin/EndDisplay.*/
	bool begin_display();
	
	/* Unlocks the drawing surface. Called after the drawing is finished.*/
	void end_display();
	
	/* Shows the result of drawing on the screen. effect=0 - normal flip,
	effect=1 - cross-fade (duration in millisec.). clearFlags specify
	whether the front or the backbuffer is to be cleared before flipping
	or fading (fade-in or fade-out).*/
	void flip (UINT effect, UINT clear_flags, UINT duration);
	
	/* Erases everything from the visible surface*/
	void clear_display();
	
	/* Sets update region - all drawing beyond this region must be ignored.*/
	void set_region (UINT x, UINT y, UINT width, UINT height);
	
	/* Blitts a piece of an image to the drawing surface.*/
	void blt (IMAGE image, UINT x, UINT y, UINT i_x, UINT i_y, UINT i_width, UINT i_height);
	
	/* As above, but stretching. */
	void stretch_blt (IMAGE image, UINT x, UINT y, UINT width, UINT height);
	
	/* Draws a vertical line. dwColor is 24-bit RGB (lsb is red). */
	void v_line (UINT x, UINT y, UINT length, DWORD color);
	
	/* Draws a horizontal line.*/
	void h_line (UINT x, UINT y, UINT length, DWORD color);
	
	/* Renders text. The return value is important when printing -
	it's the width of the printed text, given in Panorama coordinates.*/
	UINT out_text (const char* text, UINT text_len, UINT x, UINT y, DWORD color, FONT font);
	
	/* Draws a solid-filled rectangle.*/
	void fill_rect (UINT x, UINT y, UINT width, UINT height, DWORD color);
	
	/* Displays print dialog boxes and prints selected pages using
	Panorama::print_callback(). The callback uses stretch_blt(),
	v_line, h_line() and out_text() to draw particular objects.
	The SI must distinguish calls to these methods during printing
	from normal drawing calls.*/
	void print( const char* jobname );
	
	/* Makes the cursor visible or hidden. The cursor will not be drawn if the
	image wasn't set. On startup the status should be: hidden.*/
	void show_cursor (bool show);
	
	/* Moves the cursor to another location.*/
	void move_cursor (UINT x, UINT y);
	
	/* Sets new cursor image.*/
	void set_cursor (IMAGE image, UINT xspot, UINT yspot);
	
	/* Loads a sound from a buffer.*/
	SOUND load_sound (BUFFERPTR buf, DWORD size, DWORD freq, UINT flags);
	
	/* Plays a loaded sound. */
	void play_sound (SOUND sound);
	
	/* Removes a sound from memory. */
	void unload_sound (SOUND sound);
	
	/* Play a song stored in memory */
	bool play_song (const char* filename, bool loop, BUFFERPTR buf, DWORD size);
	
	/* Play a song stored in current directory */
	bool play_song (const char* filename, bool loop);
	
	/* Stop (and unload from memory) a song */
	void stop_song (bool b_fade_out);
	
	/* Pause/unpause a loaded song */
	void pause_song();
	
	/* Control overall sound volume (>0 - louder, <0 - quieter) */
	void volume_control (int direction);
	
	/* Prepares a movie for playback. Also returns framerate (fps)*/
	MOVIE load_movie (const char* filename, bool loop, UINT* fps);
	
	/* Advances movie frame */
	void update_frame (MOVIE movie);
	
	/* Displays a movie frame */
	void blt (MOVIE movie, UINT x, UINT y, UINT width, UINT height);
	
	/* Unloads the movie */
	void unload_movie (MOVIE movie);
	
	/* Gets current system time in millisec. */
	DWORD get_time();
	
	/* Starts a timer (issuing a message every "period" ms). */
	void start_timer (UINT msg, UINT period);
	
	/* Stops a timer. */
	void stop_timer (UINT msg);
	
	/* Retrieves a message (if any).*/
	bool get_message (UINT& msg, DWORD& param);
	
	/* Calls idle processing routine.*/
	void idle();
	
	/* Posts a message. */
	void post_message (UINT msg, DWORD param);
	
	/* Gets user flag*/
	UINT get_user_flag (UINT flag);
	
	/* Gets user string */
	char* get_user_string (UINT flag);
	
	/* Opens http:// mailto: ftp: or anything else*/
	void open_link (const char* text);
	
	/* Puts a string into clipboard.*/
	void put_to_clipboard (const char* text);
	
	/* Building search database*/
	bool scan (UINT progress);
	
	/* Article search dialog */
	bool search (SearchDatabase* sdb, char* &word, char* &article);
	
	/* Checks a pointer if non-NULL. If NULL - makes the system shut down.*/
	void critical_check (void* pointer);
	
	/* Sets an error message displayed upon exit. NULL if no message to be d.*/
	void set_error_message (char* message);
	
	/* Output debug message (used only in debug versions) */
	void debug_message (const char* message);

	/* Checks all font paths tofind the specified font, and returns a complete path to the first it finds */
	char* findFontPath (const char* face);
	
	char* getErrorMessage ();

	void toComponents(DWORD color, Uint8* red, Uint8* green, Uint8* blue);

	/* Search auxillary functions */
	void findWords(SearchDatabaseTree* tree, Vector* word, Vector* wordCollection);
	char* extractWord(Vector* word);
	SearchDatabaseTree* find(SearchDatabaseTree* sdb, char* word);
	void findAll(SearchDatabaseTree* sdb, char* beginsWith, Vector* words);
	void findArticles(SearchDatabase* database, char* word, Vector* articles);

	
//	Map* SDLSystem::createFontMapping(Vector* files, Map* mapping);
	Map* createFontMapping(Vector* files, Map* mapping);

	
	private:
	
	FILE* openFile;
	Platform* platform;
	Config* config;
	
	SDL_Surface *current;
	SDL_Surface *videoSurface;
	bool ready;
	Vector* messageBuffer;
	Queue* messageQueue;
	SDL_mutex* postSync;
	
	/* Data relating to the cursor */
	SDL_Surface* cursorImage;
	SDL_Rect cursorArea;
	UINT cursorX;
	UINT cursorY;
	bool cursorShow;
	
	/* Data relating to font handling */
	Map* fontBuffer;
	Map* fontMap;
	FontMapping* fontMapping;
	
	/* Data relating to debug */
	char* errorMessage;
	
	/* Data relating to searching */
	SearchPanel* searchPanel;
	Vector* searchWords;
	Vector* searchArticles;
	
	/* Data relating to sound output */
	SoundSystem* soundSystem;
	
	/* Data releating to file system access */
	Vector* filesInDir;
	Vector* foundFiles;
	int fileIndex;
	
};

#endif

